package org.bdware.server.nodecenter;

import com.google.gson.JsonObject;
import com.google.gson.JsonParser;
import io.netty.util.internal.StringUtil;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.bdware.sc.Jedion.KV;
import org.bdware.sc.conn.ByteUtil;
import org.bdware.sc.conn.ResultCallback;
import org.bdware.sc.db.KeyValueDBUtil;
import org.bdware.sc.encrypt.HardwareInfo;
import org.bdware.sc.util.JsonUtil;
import org.bdware.server.action.Action;
import org.bdware.server.permission.Role;
import org.bouncycastle.crypto.params.ECPublicKeyParameters;
import org.bouncycastle.pqc.math.linearalgebra.ByteUtils;
import org.zz.gmhelper.BCECUtil;
import org.zz.gmhelper.SM2Util;

import java.text.SimpleDateFormat;
import java.util.*;

// CenterPortal用户管理相关
public class NCManagerAction {
    // key为centerManager的数据表示平台管理员(中心管理员)
    public static final String centerManger = "__CenterManager__";
    public static final String clusterName = "__ClusterName__";
    static final String Licence = "__LICENCE___";
    static final ECPublicKeyParameters licencePub =
            BCECUtil.createECPublicKeyFromStrParameters(
                    "04844412ecb77b07d5ad7097c488ae9dff1013b310d2311f8bd84c491642011e1a6a7041bae8c2ad75da29c27e320bd430abc723cf6bcb0490afc82cc26e6d5637",
                    SM2Util.CURVE,
                    SM2Util.DOMAIN_PARAMS);
    private static final Logger LOGGER = LogManager.getLogger(NCManagerAction.class);
    String sessionID = null;
    String pubKey;
    Random random = new Random();
    private NodeCenterWSFrameHandler handler;

    public NCManagerAction(NodeCenterWSFrameHandler nodeCenterWSFrameHandler) {
        this.handler = nodeCenterWSFrameHandler;
        pubKey = null;
    }

    public static boolean isCenterManager(String pubkey) {
        String ret = KeyValueDBUtil.instance.getValue(NCTables.ConfigDB.toString(), centerManger);
        LOGGER.debug("centerManger: " + ret);
        return !StringUtil.isNullOrEmpty(ret)
                && !StringUtil.isNullOrEmpty(pubkey)
                && pubkey.equals(ret);
    }

    private void simpleReply(ResultCallback resultCallback, String action, String data) {
        JsonObject jsonResult = new JsonObject();
        jsonResult.addProperty("action", action);
        jsonResult.addProperty("data", data);
        resultCallback.onResult(jsonResult);
    }

    @Action(userPermission = 0)
    public void getNodeSessionID(JsonObject json, ResultCallback resultCallback) {
        getSessionID(json, resultCallback);
    }

    @Action(userPermission = 0)
    public void getSessionID(JsonObject json, ResultCallback resultCallback) {
        long start = System.currentTimeMillis();
        sessionID = random.nextLong() + "_session";
        JsonObject result = new JsonObject();
        result.addProperty("action", "onSessionID");
        result.addProperty("session", sessionID);
        resultCallback.onResult(result);
        long end = System.currentTimeMillis();
        LOGGER.debug("time:" + (end - start) + "data:" + JsonUtil.toJson(result));
    }

    @Action(userPermission = 0)
    public void getNodeRole(JsonObject json, ResultCallback resultCallback) {
        getRole(json, resultCallback);
    }

    @Action(userPermission = 0)
    public void getManagerPubkey(JsonObject json, ResultCallback resultCallback) {
        String ret = KeyValueDBUtil.instance.getValue(NCTables.ConfigDB.toString(), centerManger);
        simpleReply(resultCallback, "onGetManagerPubkey", ret);
    }

    @Action(userPermission = 0)
    public void getClusterName(JsonObject json, ResultCallback resultCallback) {
        String ret = KeyValueDBUtil.instance.getValue(NCTables.ConfigDB.toString(), clusterName);
        simpleReply(resultCallback, "onGetClusterName", ret);
    }

    @Action(userPermission = 0)
    public void setClusterName(JsonObject json, ResultCallback resultCallback) {
        if (json.has("name")) {
            KeyValueDBUtil.instance.setValue(
                    NCTables.ConfigDB.toString(), clusterName, json.get("name").getAsString());
            simpleReply(resultCallback, "onSetClusterName", "success");
        } else {
            simpleReply(resultCallback, "onSetClusterName", "failed");
        }
    }

    public void getRole(JsonObject json, ResultCallback resultCallback) {
        if (pubKey == null) {
            simpleReply(resultCallback, "onLogin", Role.Anonymous.name());
            return;
        }
        try {
            String ret =
                    KeyValueDBUtil.instance.getValue(NCTables.ConfigDB.toString(), centerManger);
            if (ret != null && ret.length() > 0) {
                boolean isCenterManager = (ret.equals(pubKey)); // 表示此节点是否是平台管理员
                ret = KeyValueDBUtil.instance.getValue(NCTables.NodeUser.toString(), pubKey);
                String role = "";
                if (isCenterManager) {
                    role = "CenterManager,";
                }
                if (null != ret && !ret.isEmpty()) {
                    role += ret;
                }
                if (role.endsWith(",")) {
                    role = role.substring(0, role.length() - 1);
                } else if (role.isEmpty()) {
                    role = Role.Anonymous.name();
                }
                handler.setPermission(Role.compoundValue(role.split(",")));
                simpleReply(resultCallback, "onLogin", role);
            } else {
                KeyValueDBUtil.instance.setValue(
                        NCTables.ConfigDB.toString(), centerManger, pubKey);
                KeyValueDBUtil.instance.setValue(
                        NCTables.ConfigDB.toString(), clusterName, "clusterName_" + pubKey.substring(0, 5));
                handler.setPermission(0x30000ffL);
                simpleReply(resultCallback, "onLogin", "CenterManager");
            }
        } catch (Exception e) {
            e.printStackTrace();
            simpleReply(resultCallback, "onLogin", Role.Anonymous.name());
        }
    }

    @Action(userPermission = 0)
    public void login(JsonObject json, ResultCallback resultCallback) {
        long start = System.currentTimeMillis();
        JsonObject jsonResult = new JsonObject();
        if (sessionID == null) {
            simpleReply(resultCallback, "onLogin", Role.Anonymous.name());
            return;
        }
        String signature = json.get("signature").getAsString();
        String pubKey = json.get("pubKey").getAsString();

        boolean result = SM2Util.plainStrVerify(pubKey, sessionID, signature);
        LOGGER.debug("session:" + (sessionID));
        if (result) {
            this.pubKey = pubKey;
            LOGGER.debug("设置公钥" + pubKey);
            getRole(json, resultCallback);
        } else {
            simpleReply(resultCallback, "onLogin", Role.Anonymous.name());
        }
        long end = System.currentTimeMillis();
        LOGGER.debug("time:" + (end - start));
    }

    @Action(userPermission = 0)
    public void applyRole(JsonObject json, ResultCallback resultCallback) {
        long start = System.currentTimeMillis();
        if (pubKey == null) {
            simpleReply(resultCallback, "onApplyRole", "missing pubKey");
            return;
        }
        if (json.has("role")) {
            Role role = Role.parse(json.get("role").getAsString());
            if (role != Role.Anonymous) {
                applyAtDB(role, resultCallback, start);
                return;
            }
        }
        simpleReply(resultCallback, "onApplyRole", "failed");
    }

    @Action(userPermission = 1 << 6)
    public void addNode(JsonObject json, ResultCallback resultCallback) {
        try {
            KeyValueDBUtil.instance.setValue(
                    NCTables.NodeUser.toString(),
                    json.get("nodePubKey").getAsString(),
                    Role.Node.toString());
            KeyValueDBUtil.instance.setValue(
                    NCTables.NodeTime.toString(),
                    json.get("nodePubKey").getAsString(),
                    Long.toString(new Date().getTime()));
            simpleReply(resultCallback, "onAddNodes", "success");
        } catch (Exception e) {
            e.printStackTrace();
            simpleReply(resultCallback, "onAddNodes", "failed");
        }
    }

    private void applyAtDB(Role role, ResultCallback resultCallback, long start) {
        String str = KeyValueDBUtil.instance.getValue(NCTables.ApplyRole.toString(), pubKey);
        String already = KeyValueDBUtil.instance.getValue(NCTables.NodeUser.toString(), pubKey);
        if (already != null && already.contains(role.toString())) {
            simpleReply(resultCallback, "onApplyRole", "already has!");
            return;
        }
        if (str == null || str.length() == 0) {
            KeyValueDBUtil.instance.setValue(NCTables.ApplyRole.toString(), pubKey, role.name());
            KeyValueDBUtil.instance.setValue(
                    NCTables.ApplyTime.toString(), pubKey, Long.toString(new Date().getTime()));
        } else {
            if (!str.contains(role.name())) {
                KeyValueDBUtil.instance.setValue(
                        NCTables.ApplyRole.toString(), pubKey, str + "," + role.name());
                KeyValueDBUtil.instance.setValue(
                        NCTables.ApplyTime.toString(), pubKey, Long.toString(new Date().getTime()));
            }
        }
        simpleReply(resultCallback, "onApplyRole", "success!");
    }

    @Action(userPermission = 1 << 3)
    public void listAllUsers(JsonObject json, ResultCallback resultCallback) {
        List<KV> kv = KeyValueDBUtil.instance.getKeyValues(NCTables.NodeUser.toString());
        List<KV> time = KeyValueDBUtil.instance.getKeyValues(NCTables.NodeTime.toString());
        Map<String, Object> ret = new HashMap<>();
        ret.put("kv", kv);
        ret.put("time", time);
        Map<String, Object> result = new HashMap<>();
        result.put("action", "onListAllUsers");
        result.put("data", ret);
        resultCallback.onResult(result);
    }

    @Action(userPermission = 1 << 2)
    public void authNodeManager(JsonObject json, ResultCallback resultCallback) {
        long start = System.currentTimeMillis();
        String pubKey = json.get("pubKey").getAsString();
        boolean isAccept = json.get("isAccept").getAsBoolean();
        LOGGER.debug("[NCManagerAction] " + json.toString());
        if (pubKey == null || pubKey.length() == 0) {
            simpleReply(resultCallback, "onAuthNodeManager", "missing pubKey");
            return;
        }
        if (!isAccept) {
            KeyValueDBUtil.instance.delete(NCTables.ApplyRole.toString(), pubKey);
            KeyValueDBUtil.instance.delete(NCTables.ApplyTime.toString(), pubKey);
            simpleReply(resultCallback, "onAuthNodeManager", "success");
            return;
        }
        String already = KeyValueDBUtil.instance.getValue(NCTables.NodeUser.toString(), pubKey);
        String roles = KeyValueDBUtil.instance.getValue(NCTables.ApplyRole.toString(), pubKey);
        if (roles == null || roles.length() == 0) {
            KeyValueDBUtil.instance.delete(NCTables.ApplyRole.toString(), pubKey);
            KeyValueDBUtil.instance.delete(NCTables.ApplyTime.toString(), pubKey);
            simpleReply(resultCallback, "onAuthNodeManager", "empty apply!");
            return;
        }
        if (already != null && already.length() > 0) {
            already += "," + roles;
        } else {
            already = roles;
        }
        KeyValueDBUtil.instance.setValue(NCTables.NodeUser.toString(), pubKey, already);
        KeyValueDBUtil.instance.setValue(
                NCTables.NodeTime.toString(), pubKey, Long.toString(new Date().getTime()));
        KeyValueDBUtil.instance.delete(NCTables.ApplyRole.toString(), pubKey);
        KeyValueDBUtil.instance.delete(NCTables.ApplyTime.toString(), pubKey);
        simpleReply(resultCallback, "onAuthNodeManager", "success");
        long end = System.currentTimeMillis();
        LOGGER.debug("[authNodeManager:time]" + (end - start));
    }

    @Action(userPermission = 1 << 4)
    public void listApplyList(JsonObject json, ResultCallback resultCallback) {
        long start = System.currentTimeMillis();
        List<KV> kv = KeyValueDBUtil.instance.getKeyValues(NCTables.ApplyRole.toString());
        List<KV> time = KeyValueDBUtil.instance.getKeyValues(NCTables.ApplyTime.toString());
        Map<String, Object> ret = new HashMap<>();
        ret.put("kv", kv);
        ret.put("time", time);
        Map<String, Object> result = new HashMap<>();
        result.put("action", "onListApplyList");
        result.put("data", ret);
        resultCallback.onResult(result);
        long end = System.currentTimeMillis();
        LOGGER.debug("[listApplyList:time]" + (end - start));
    }

    @Action(userPermission = 1 << 5)
    public void delete(JsonObject json, ResultCallback resultCallback) {
        String pubKey = json.get("pubKey").getAsString();
        KeyValueDBUtil.instance.delete(NCTables.NodeUser.toString(), pubKey);
        KeyValueDBUtil.instance.delete(NCTables.NodeTime.toString(), pubKey);
        simpleReply(resultCallback, "onDelete", "success");
    }

    @Action(userPermission = 1 << 4)
    public void listLicence(JsonObject json, ResultCallback resultCallback) {
        long start = System.currentTimeMillis();
        String licence = KeyValueDBUtil.instance.getValue(NCTables.ConfigDB.toString(), Licence);
        if (licence == null || licence.length() == 0) {
            JsonObject jsonResult = new JsonObject();
            jsonResult.addProperty("action", "onListLicence");
            jsonResult.addProperty("nodeCount", 0);
            jsonResult.addProperty("dueDate", "已过期");
            resultCallback.onResult(jsonResult);
            return;
        }
        json = JsonParser.parseString(licence).getAsJsonObject();
        String sign = json.get("sign").getAsString();
        String content = json.get("content").getAsString();
        if (StringUtil.isNullOrEmpty(sign) || StringUtil.isNullOrEmpty(content)) {
            JsonObject jsonResult = new JsonObject();
            jsonResult.addProperty("action", "onListLicence");
            jsonResult.addProperty("nodeCount", 0);
            jsonResult.addProperty("dueDate", "已过期");
            resultCallback.onResult(jsonResult);
            return;
        }
        long expiredDate;
        long nodeCount;
        try {
            JsonObject jo = JsonParser.parseString(content).getAsJsonObject();
            nodeCount = jo.get("nodeCount").getAsLong();
            expiredDate = jo.get("expiredDate").getAsLong();
        } catch (Exception e) {
            JsonObject jsonResult = new JsonObject();
            jsonResult.addProperty("action", "onListLicence");
            jsonResult.addProperty("nodeCount", 0);
            jsonResult.addProperty("dueDate", "已过期");
            resultCallback.onResult(jsonResult);
            return;
        }
        if (expiredDate == 0 || nodeCount == 0) {
            JsonObject jsonResult = new JsonObject();
            jsonResult.addProperty("action", "onListLicence");
            jsonResult.addProperty("nodeCount", 0);
            jsonResult.addProperty("dueDate", "已过期");
            resultCallback.onResult(jsonResult);
            return;
        }
        SimpleDateFormat format = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
        boolean verify =
                SM2Util.verify(licencePub, content.getBytes(), ByteUtils.fromHexString(sign));
        if (verify) {
            JsonObject jsonResult = new JsonObject();
            jsonResult.addProperty("action", "onListLicence");
            jsonResult.addProperty("nodeCount", nodeCount);
            jsonResult.addProperty("dueDate", format.format(new Date(expiredDate)));
            resultCallback.onResult(jsonResult);
            long end = System.currentTimeMillis();
            LOGGER.debug("[listLicence:time]" + (end - start));
        } else {
            JsonObject jsonResult = new JsonObject();
            jsonResult.addProperty("action", "onListLicence");
            jsonResult.addProperty("nodeCount", 0);
            jsonResult.addProperty("dueDate", "已过期");
            resultCallback.onResult(jsonResult);
            long end = System.currentTimeMillis();
            LOGGER.debug("[listLicence:time]" + (end - start));
        }
    }

    @Action(userPermission = 1L << 2L)
    public void updateLicence(JsonObject json, ResultCallback resultCallback) {
        long start = System.currentTimeMillis();
        String sign = json.get("sign").getAsString();
        String content = json.get("content").getAsString();
        if (StringUtil.isNullOrEmpty(sign) || StringUtil.isNullOrEmpty(content)) {
            simpleReply(resultCallback, "onUpdateLicence", "missing signature or content");
            return;
        }
        long expiredDate;
        long nodeCount;
        String uuid;
        try {
            JsonObject jo = JsonParser.parseString(content).getAsJsonObject();
            nodeCount = jo.get("nodeCount").getAsLong();
            expiredDate = jo.get("expiredDate").getAsLong();
            uuid = jo.get("uuid").getAsString();

        } catch (Exception e) {
            simpleReply(resultCallback, "onUpdateLicence", "content error");
            return;
        }
        if (expiredDate == 0 || nodeCount == 0) {
            simpleReply(resultCallback, "onUpdateLicence", "parse expiredDate/nodeCount error");
            return;
        }
        if (!uuid.equals(
                ByteUtil.encodeBASE64(HardwareInfo.getCPUID().getBytes()).replaceAll("\n", ""))) {
            simpleReply(resultCallback, "onUpdateLicence", "invalid licence");
            return;
        }
        boolean verify =
                SM2Util.verify(licencePub, content.getBytes(), ByteUtils.fromHexString(sign));

        if (verify) {
            KeyValueDBUtil.instance.setValue(
                    NCTables.ConfigDB.toString(), Licence, json.toString());
            simpleReply(resultCallback, "onUpdateLicence", "success");
            long end = System.currentTimeMillis();
            LOGGER.debug("[listLicence:time]" + (end - start));
            return;
        } else simpleReply(resultCallback, "onUpdateLicence", "failed");
        long end = System.currentTimeMillis();
        LOGGER.debug("[updateLicence:time]" + (end - start));
    }

    @Action(userPermission = 1 << 9)
    public void getOtherNC(JsonObject json, ResultCallback resultCallback) {
        Map<String, String> result = new HashMap<>();
        result.put("action", "onGetOtherNC");
        String s = OtherNCProxy.instance.getOtherNCs().replace(";", " ");
        result.put("address", s);
        resultCallback.onResult(result);
    }

    // 改为NodeManager权限
    @Action(userPermission = 1 << 9)
    public void changeOtherNC(JsonObject json, ResultCallback resultCallback) {
        String add = json.get("data").getAsString().replace(" ", ";");
        String s = OtherNCProxy.instance.setOtherNCs(add).replace(";", " ");
        JsonObject result = new JsonObject();
        result.addProperty("action", "onGetOtherNC");
        result.addProperty("address", s);
        resultCallback.onResult(result);
    }

    // NodeManager权限
    @Action(userPermission = 1 << 9)
    public void changeNCFile(JsonObject json, ResultCallback resultCallback) {
        String add = json.get("data").getAsString();
        String s = OtherNCProxy.instance.setNCFile(add);
        JsonObject result = new JsonObject();
        result.addProperty("action", "onGetNCFile");
        result.addProperty("fileName", s);
        resultCallback.onResult(result);
    }

    @Action(userPermission = 1 << 9)
    public void getNCFile(JsonObject json, ResultCallback resultCallback) {
        String s = OtherNCProxy.instance.getNCFile();
        Map<String, String> result = new HashMap<>();
        result.put("action", "onGetNCFile");
        result.put("fileName", s);
        resultCallback.onResult(result);
    }

    // TODO 迁移合约实例（从节点A将合约M的实例迁移到节点B）的权限?
    /*
     * now only support Sole contract type
     */
    @Action(userPermission = 0)
    public void transferContractInstance(JsonObject json, ResultCallback resultCallback) {
        LOGGER.info("transferContractInstance");
        String node1ID = json.get("node1ID").getAsString();
        String node2ID = json.get("node2ID").getAsString();
        String contractID = json.get("contractID").getAsString();

        CMNode node1 = NodeCenterActions.nodeInfos.get(node1ID);
        if (node1 == null) {
            LOGGER.info("node1 " + node1ID + " is null!");
            return;
        }
        if (node2ID == null) {
            LOGGER.info("node2 " + node1ID + " is null!");
            return;
        }
        if (node1ID.equals(node2ID)) {
            LOGGER.info("node1 and node2 is the same node.");
            return;
        }

        Map<String, String> request = new HashMap<>();
        request.put("action", "transferInstance");
        request.put("contractID", contractID);
        request.put("nodeID", node2ID);
        CMNode node2 = NodeCenterActions.nodeInfos.get(node2ID);
        if (node2 == null) {
            LOGGER.info("node  " + node2ID + " is null!");
            return;
        }
        request.put("address", node2.masterAddress);
        node1.connection.controller.sendMsg(JsonUtil.toJson(request));

        Map<String, String> result = new HashMap<>();
        result.put("action", "onTransfer");
        result.put(
                "data",
                "now start transfer contract "
                        + contractID
                        + " from node:"
                        + node1ID
                        + " to node:"
                        + node2ID);
        resultCallback.onResult(result);
    }
}
